/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base;

import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.Loadable;
import java.nio.IntBuffer;

/**
 * Used to draw a primitive with some options.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class Mesh extends Loadable {
    
    /**
     * The primitive to draw.
     */
    private final Primitive primitive;
    
    /**
     * Which shape to draw the primitive.
     */
    private final DrawingType drawingType;
    
    /**
     * How many primitives to draw.
     */
    private int instances;
    
    /**
     * The size of each pixels in a POINT drawing type.
     */
    private float pointSize;
    
    /**
     * If the back faces are hiden or not (by default true).
     */
    private boolean faceCulling;
    
    /**
     * Create a mesh with a primitive and a drawing type.
     * @param primitive Primitive to draw.
     * @param drawingType Which shape to draw the primitive.
     */
    public Mesh(final Primitive primitive, final DrawingType drawingType) {
        this.primitive = primitive;
        this.drawingType = drawingType;
        this.instances = 1;
        this.pointSize = 4f;
        this.faceCulling = true;
    }

    /**
     * Get the primitive attached with this mesh.
     * @return Primitive to draw.
     */
    public final Primitive getPrimitive() {
        return primitive;
    }

    /**
     * Specify how many primitives to draw.
     * @param instances 
     */
    public final void setInstances(int instances) {
        this.instances = instances;
    }

    /**
     * Change the size of each pixels in a POINT drawing type.
     * @param pointSize 
     */
    public final void setPointSize(final float pointSize) {
        this.pointSize = pointSize;
    }

    /**
     * If the back faces are hiden or not (by default true).
     * @param faceCulling 
     */
    public final void setFaceCulling(final boolean faceCulling) {
        this.faceCulling = faceCulling;
    }

    @Override
    protected final void onLoad() {        
        if (primitive != null) {
            primitive.load();
        }
    }

    @Override
    protected final void onBind() {
        /* Get opengl 3.0 object */
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
                
        if (primitive != null) {
            primitive.bind();
            
            /* Change the size of each pixels in a POINT drawing type */
            gl.glPointSize(pointSize);
            
            /* Hide back faces if face culling is enabled */
            if (faceCulling) {
                gl.glCullFace(GL3bc.GL_BACK);
                gl.glEnable(GL3bc.GL_CULL_FACE);
            } else {
                gl.glDisable(GL3bc.GL_CULL_FACE);
            }
            
            /* Be sure the selected drawing mode can be drawed */
            DrawingType finalDrawingType = drawingType;
            
//            if (finalDrawingType == DrawingType.POINTS && !primitive.hasPointFaces()) {
//                finalDrawingType = DrawingType.LINES;
//            }
//            if (finalDrawingType == DrawingType.LINES && !primitive.hasLineFaces()) {
//                finalDrawingType = DrawingType.TRIANGLES;
//            }
//            if (finalDrawingType == DrawingType.TRIANGLES && !primitive.hasTriangleFaces()) {
//                finalDrawingType = DrawingType.QUADS;
//            }
//            if (finalDrawingType == DrawingType.QUADS && !primitive.hasQuadFaces()) {
//                finalDrawingType = null;
//            }
            
            /* Draw the available and selected drawing mode */
            if (null != finalDrawingType) {
                switch (finalDrawingType) {
                    case POINTS:
                        /* Get vertex positions */
                        final IntBuffer pointsBuffer = 
                                primitive.getPointIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_POINTS, pointsBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, pointsBuffer, instances);
                        break;
                    case LINES:
                        /* Get vertex positions */
                        final IntBuffer linesBuffer =
                                primitive.getLineIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_LINES, linesBuffer.limit(),
                               GL3bc.GL_UNSIGNED_INT, linesBuffer, instances);
                        break;
                    case TRIANGLES:
                        /* Get vertex positions */
                        final IntBuffer trianglesBuffer =
                                primitive.getTriangleIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_TRIANGLES, trianglesBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, trianglesBuffer, instances);
                        break;
                    case QUADS:
                        /* Get vertex positions */
                        final IntBuffer quadsBuffer =
                                primitive.getQuadIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_QUADS, quadsBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, quadsBuffer, instances);
                        break;
                }
            }
        }
    }

    @Override
    protected final void onUnload() {
        if (primitive != null) {
            primitive.unload();
        }
    }
    
}
