/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.Loadable;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.LineFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.PointFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.QuadFace;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.faces.TriangleFace;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.LinkedList;
import java.util.List;

/**
 * Used to store vertices and faces to draw a 3D object.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 * @see Face
 * @see DrawingType
 */
public abstract class Primitive extends Loadable {

    /**
     * The list of vertices.
     */
    private final List<Vertex> vertices;
    
    /**
     * The list of POINT drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final List<PointFace> pointFaces;
    
    /**
     * The list of LINE drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final List<LineFace> lineFaces;
    
    /**
     * The list of triangles drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final List<TriangleFace> triangleFaces;
    
    /**
     * The list of quad (square) drawing mode faces.
     * @see Face
     * @see DrawingType
     */
    private final List<QuadFace> quadFaces;

    /**
     * Vertex array object.
     */
    private final IntBuffer vao;
    
    /**
     * Vertex buffer object.
     */
    private final IntBuffer vbo;

    /**
     * Vertices buffer.
     */
    private FloatBuffer vertexPositionBuffer;
    
    /**
     * Point faces buffer.
     */
    private IntBuffer pointsBuffer;
    
    /**
     * Line faces buffer.
     */
    private IntBuffer linesBuffer;
    
    /**
     * Triangle faces buffer.
     */
    private IntBuffer trianglesBuffer;
    
    /**
     * Quad (square) faces buffer.
     */
    private IntBuffer quadsBuffer;

    /**
     * Create an empty primitive object.
     */
    public Primitive() {
        this.vertices = new LinkedList();
        this.pointFaces = new LinkedList();
        this.lineFaces = new LinkedList();
        this.triangleFaces = new LinkedList();
        this.quadFaces = new LinkedList();

        this.vao = Buffers.newDirectIntBuffer(1);
        this.vbo = Buffers.newDirectIntBuffer(1);

        this.vertexPositionBuffer = Buffers.newDirectFloatBuffer(0);
        this.pointsBuffer = Buffers.newDirectIntBuffer(0);
        this.linesBuffer = Buffers.newDirectIntBuffer(0);
        this.trianglesBuffer = Buffers.newDirectIntBuffer(0);
        this.quadsBuffer = Buffers.newDirectIntBuffer(0);
    }

    /**
     * Add a vertex (use-it in the generateVertices function)
     * @param vertex 
     */
    public final void addVertex(final Vertex vertex) {
        vertices.add(vertex);
    }

    /**
     * Add a point face (use-it in the generatePointFace function)
     * @param face 
     */
    public final void addPointFace(final PointFace face) {
        pointFaces.add(face);
    }

    /**
     * Add a line face (use-it in the generateLineFace function)
     * @param face 
     */
    public final void addLineFace(final LineFace face) {
        lineFaces.add(face);
    }

    /**
     * Add a triangle face (use-it in the generateTriangleFace function)
     * @param face 
     */
    public final void addTriangleFace(final TriangleFace face) {
        triangleFaces.add(face);
    }

    /**
     * Add a quad (square) face (use-it in the generateQuadFace function)
     * @param face 
     */
    public final void addQuadFace(final QuadFace face) {
        quadFaces.add(face);
    }

    /**
     * Get the number of vertex.
     * Can be called meny times.
     * @return 
     */
    public final int getVertexCount() {
        if (vertices.isEmpty()) {
            generateVertices();
        }

        return vertices.size();
    }

    /**
     * Get the vertices buffer.
     * Can be called meny times.
     * @return 
     */
    public final FloatBuffer getVertexPositionBuffer() {
        if (vertices.isEmpty()) {
            generateVertices();
        }

        if (vertexPositionBuffer.limit() == 0) {
            final float[] positions = new float[vertices.size() * 3];

            int iteration = 0;

            for (final Vertex vertex : vertices) {
                if (vertex != null) {
                    if (vertex.getPosition() != null) {
                        positions[iteration] = vertex.getPosition().x;
                        iteration++;
                        positions[iteration] = vertex.getPosition().y;
                        iteration++;
                        positions[iteration] = vertex.getPosition().z;
                        iteration++;
                    }
                }
            }

            vertexPositionBuffer = Buffers.newDirectFloatBuffer(positions);
        }

        return vertexPositionBuffer;
    }

    /**
     * Implements-it to generate each vertex.
     */
    protected abstract void generateVertices();

    /**
     * Get if this primitive has line faces or not.
     * @return 
     */
    public final boolean hasPointFaces() {
        return !pointFaces.isEmpty();
    }
    
    /**
     * Get the point index buffer.
     * Can be called meny times.
     * @return 
     */
    public final IntBuffer getPointIndexBuffer() {
        if (!hasPointFaces()) {
            generatePointFaces();

            final int[] indices = new int[pointFaces.size()];

            int iteration = 0;

            for (final Face face : pointFaces) {
                if (face != null) {
                    for (final int index : face.getIndices()) {
                        indices[iteration] = index;
                        iteration++;
                    }
                }
            }

            pointsBuffer = Buffers.newDirectIntBuffer(indices);
        }

        return pointsBuffer;
    }

    /**
     * Implement-it to generate point faces.
     */
    protected abstract void generatePointFaces();

    /**
     * Get if this primitive has line faces or not.
     * @return 
     */
    public final boolean hasLineFaces() {
        return !lineFaces.isEmpty();
    }
    
    /**
     * Get the line index buffer.
     * Can be called meny times.
     * @return 
     */
    public final IntBuffer getLineIndexBuffer() {
        if (!hasLineFaces()) {
            generateLineFaces();

            final int[] indices = new int[lineFaces.size() * 2];

            int iteration = 0;

            for (final Face face : lineFaces) {
                if (face != null) {
                    for (final int index : face.getIndices()) {
                        indices[iteration] = index;
                        iteration++;
                    }
                }
            }

            linesBuffer = Buffers.newDirectIntBuffer(indices);
        }

        return linesBuffer;
    }

    /**
     * Implements it to generate line faces.
     */
    protected abstract void generateLineFaces();

    /**
     * Get if this primitive has triangle faces or not.
     * @return 
     */
    public final boolean hasTriangleFaces() {
        return !triangleFaces.isEmpty();
    }
    
    /**
     * Get the triangle index buffer.
     * Can be called many times.
     * @return 
     */
    public final IntBuffer getTriangleIndexBuffer() {
        if (!hasTriangleFaces()) {
            generateTriangleFaces();

            final int[] indices = new int[triangleFaces.size() * 3];

            int iteration = 0;

            for (final Face face : triangleFaces) {
                if (face != null) {
                    for (final int index : face.getIndices()) {
                        indices[iteration] = index;
                        iteration++;
                    }
                }
            }

            trianglesBuffer = Buffers.newDirectIntBuffer(indices);
        }

        return trianglesBuffer;
    }

    /**
     * Implement it to generate triangle faces.
     */
    protected abstract void generateTriangleFaces();

    /**
     * Get if this primitive has quad faces or not.
     * @return 
     */
    public final boolean hasQuadFaces() {
        return !quadFaces.isEmpty();
    }
    
    /**
     * Get the quad index buffer.
     * Can be called many times.
     * @return 
     */
    public final IntBuffer getQuadIndexBuffer() {
        if (!hasQuadFaces()) {
            generateQuadFaces();

            final int[] indices = new int[quadFaces.size() * 4];

            int iteration = 0;

            for (final Face face : quadFaces) {
                if (face != null) {
                    for (final int index : face.getIndices()) {
                        indices[iteration] = index;
                        iteration++;
                    }
                }
            }

            quadsBuffer = Buffers.newDirectIntBuffer(indices);
        }

        return quadsBuffer;
    }

    /**
     * Implement it to generate quad faces.
     */
    protected abstract void generateQuadFaces();

    @Override
    protected final void onLoad() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (getVertexCount() > 0) {
            /* Create vertex arrays */
            gl.glGenVertexArrays(vao.limit(), vao);
            gl.glBindVertexArray(vao.get(0));

            /* Create vertex buffers */
            gl.glGenBuffers(vbo.limit(), vbo);

            /* Set vertex positions buffer */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    getVertexPositionBuffer().limit() * Float.BYTES,
                    getVertexPositionBuffer(), GL3bc.GL_STATIC_DRAW);
        }
    }

    @Override
    protected final void onBind() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

        if (getVertexCount() > 0) {
            /* Bind vertex arrays */
            gl.glBindVertexArray(vao.get(0));

            /* Enable vertex position attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            gl.glEnableVertexAttribArray(0);
            gl.glVertexAttribPointer(0, 3, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }

    @Override
    protected final void onUnload() {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

        if (getVertexCount() > 0) {
            /* Clean up */
            gl.glDeleteBuffers(vbo.capacity(), vbo);
            gl.glDeleteVertexArrays(vao.capacity(), vao);
        }
    }

}
